from ast import literal_eval
from collections import defaultdict
from pathlib import Path
from typing import Dict, List

import numpy as np
from sklearn.metrics import confusion_matrix  # type: ignore

from pycarus.utils import get_conf_matrix_img


def _get_conf_mat_from_file(pred_file: Path) -> np.ndarray:
    """Compute confusion matrix from a prediction file.

    Args:
        pred_file: The path to the prediction file.

    Returns:
        The confusion matrix.
    """
    with open(pred_file) as f:
        items = [line.strip().split(";") for line in f.readlines()]

    labels = np.array([float(item[0]) for item in items])
    logits = np.array([literal_eval(item[1]) for item in items])
    predictions = np.argmax(logits, axis=1)

    return confusion_matrix(y_true=labels, y_pred=predictions)


def pcd_classification_evaluation(classes: List[str], pred_files: List[Path], outdir: Path) -> None:
    """Evaluation for point cloud classification.

    This function aggregates results reported in prediction files to compute:
    - class accuracies (mean and std)
    - mean class accuracy (mean and std)
    - overall accuracy (mean and std)
    - confusion matrix (mean)

    Each prediction file is expected to be produced by a different run
    and should be a txt file with one line for each prediction,
    formatted as "LABEL;LOGITS", where LABEL is a single integer and
    LOGITS is the comma separated sequence of predictions logits (floats),
    with "[" as first character and "]" as last character.
    For instance, a row could be something like:
        10;[0.1, 1.23, ..., 0.001]

    The results of the evaluation are saved in "outdir".
    In particular, the following elements will be saved:
    - results.csv, with class accuracies, mean class accuracy and overall accuracy
    - conf_matrix.png, with the confusion matrix

    Args:
        classes: A list with all the classes. The order in the list is assumed to
                 to be coherent with the labels reported in "predictions_file".
        pred_files: The list of file with all the predictions.
        outdir: The path to the directory where results will be saved.
    """
    outdir.mkdir(parents=True, exist_ok=True)

    conf_mats = [_get_conf_mat_from_file(pred_file) for pred_file in pred_files]

    num_samples: List[int] = []
    mean_accs: List[float] = []
    class_accs: Dict[int, List[float]] = defaultdict(list)
    for conf_mat in conf_mats:
        num_samples = np.sum(conf_mat, axis=1).tolist()
        class_acc = np.diag(conf_mat) / np.sum(conf_mat, axis=1)  # type: ignore
        mean_accs.append(float(np.nanmean(class_acc)))  # type: ignore

        for c in range(len(classes)):
            class_accs[c].append(float(class_acc[c]))

    results_file = outdir / "results.csv"

    with open(results_file, "wt") as f:
        f.write("Class,ID,# Samples,Accuracy,std\n")
        for c, class_name in enumerate(classes):
            accs = class_accs[c]
            if len(accs) > 1:
                mean = np.mean(np.array(accs))
                std = np.std(np.array(accs))
                f.write(f"{class_name},{c},{num_samples[c]},{mean * 100:.3f},{std * 100:.3f}\n")
            else:
                f.write(f"{class_name},{c},{num_samples[c]},{accs[0] * 100:.3f},-\n")

        f.write("\n")

        if len(mean_accs) > 1:
            mean = np.mean(np.array(mean_accs))
            std = np.std(np.array(mean_accs))
            f.write(f",,Mean,{mean * 100:.3f},{std * 100:.3f}\n")
        else:
            f.write(f",,Mean,{mean_accs[0] * 100:.3f},-\n")

        f.write("\n")

        overall_accs = [float(np.sum(np.diag(m)) / np.sum(m)) for m in conf_mats]  # type: ignore

        if len(mean_accs) > 1:
            mean = np.mean(np.array(overall_accs))
            std = np.std(np.array(overall_accs))
            f.write(f",,Overall,{mean * 100:.3f},{std * 100:.3f}\n")
        else:
            f.write(f",,Overall,{overall_accs[0] * 100:.3f},-\n")

    conf_mat_img = get_conf_matrix_img(conf_mats[0])
    conf_mat_img.save(outdir / "conf_matrix.png")



import os
from importlib.util import find_spec
from typing import Tuple

import torch

from pycarus.geometry.pcd import batchify, unbatchify


def emd(
    prediction: torch.Tensor, groundtruth: torch.Tensor, eps: float, iterations: int
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Compute the Earth Mover's Distance

    This function uses the auction algorithm for approximation, hence the assignment is not
    guranteed to be a bijection.

    Args:
        prediction: The predicted point cloud with shape ([B,] NUM_POINTS, 3). NUM_POINTS should be
        a multiple of 1024.
        groundtruth: The groundtruth point cloud with shape ([B,] NUM_POINTS, 3). NUM_POINTS should
        be a multiple of 1024.
        eps: the balance between the error rate and the speed of convergence.
        iterations: the max number of iterations.
    """
    if find_spec("emd") is None:
        folder = os.path.dirname(os.path.realpath(__file__))
        str_exp = f"EMD not found. Install it running {folder}/external/setup_emd.py"
        raise ModuleNotFoundError(str_exp)
    else:
        from pycarus.metrics.external.emd_module import EMDDistance  # type: ignore

    batched, [pred, gt] = batchify([prediction, groundtruth], 3)
    emd = EMDDistance()

    distance, assignment = emd(pred, gt, eps, iterations)

    if batched:
        [distance, assignment] = unbatchify([distance, assignment])

    return distance, assignment



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pycarus',
 'pycarus.benchmarks',
 'pycarus.benchmarks.pcd',
 'pycarus.datasets',
 'pycarus.geometry',
 'pycarus.learning',
 'pycarus.learning.models',
 'pycarus.metrics',
 'pycarus.metrics.external',
 'pycarus.transforms']

package_data = \
{'': ['*']}

install_requires = \
['einops>=0.3.0,<0.4.0',
 'h5py>=3.1.0,<4.0.0',
 'ninja>=1.10.0,<2.0.0',
 'open3d>=0.12.0,<0.13.0',
 'pykdtree>=1.3.4,<2.0.0',
 'pytorch3d>=0.3.0,<0.4.0',
 'torch>=1.7.1,<2.0.0',
 'types-setuptools>=57.0.0,<58.0.0']

setup_kwargs = {
    'name': 'pycarus',
    'version': '0.3.5',
    'description': 'Utilities for computer vision and 3D geometry',
    'long_description': None,
    'author': 'Luca De Luigi',
    'author_email': 'lucadeluigi91@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<3.9.0',
}


setup(**setup_kwargs)



from typing import Tuple

import numpy as np
import torch
from pykdtree.kdtree import KDTree  # type: ignore


def fscore(
    prediction: torch.Tensor, groundtruth: torch.Tensor, threshold: float = 0.001
) -> Tuple[float, float, float]:
    """Compute the F1-Score using the treshold as defined in:

    Knapitsch, A., Park, J., Zhou, Q. Y., & Koltun, V. (2017).
    Tanks and temples: Benchmarking large-scale scene reconstruction.
    ACM Transactions on Graphics (ToG), 36(4), 1-13.
    The function uses KdTree to compute the nearest neighbors

    Args:
        prediction: The predicted point cloud with shape (NUM_POINTS, 3).
        groundtruth: The groundtruth point cloud with shape (NUM_POINTS, 3).
        threshold: The euclidean distance treshold to use. Defaults to 0.001.

    Returns:
        A tuple with: the fscore, the precision and the recall.
    """
    pred = prediction.numpy()
    gt = groundtruth.numpy()

    kd_tree = KDTree(gt)
    dist_precision, _ = kd_tree.query(pred, k=1)

    kd_tree = KDTree(pred)
    dist_recall, _ = kd_tree.query(gt, k=1)

    fscore, recall, precision = 0.0, 0.0, 0.0

    if len(dist_precision) and len(dist_recall):
        precision = np.sum(dist_precision < threshold) / len(pred)
        recall = np.sum(dist_recall < threshold) / len(gt)

        if recall + precision > 0:
            fscore = 2 * recall * precision / (recall + precision)

    return fscore, precision, recall
